#include <QtGui/QPainter>
#include <QtCore>
#include <math.h>
#include "data.h"

Data::Data(QObject *parent) : QObject(parent) {

  mandelWatcher = 0;
  image = new QImage(imageWidth, imageHeight, QImage::Format_RGB32);
  image->fill(0xA0A0A0);
  initPalette();
}

Data::~Data() {

  if (mandelWatcher->isRunning()) {  
    mandelWatcher->cancel();
    mandelWatcher->waitForFinished();
  }  
  delete image;
  if (mandelWatcher) delete mandelWatcher;
}

void Data::initPalette() {

  int i;
  
  for (i = 0; i <= maxIterations; i++) {
    palette[i] = QColor(128.0 + 127.0 * sin(3.0 + 21.0 * (double)i / (double)maxIterations),  
                        128.0 + 127.0 * sin(17.0 * (double)i / (double)maxIterations),  
                        128.0 + 127.0 * sin(53.0 * (double)i / (double)maxIterations));
  }
}

QImage * Data::getImage() {

  return(image);
}

QSize Data::getImageSize() {

  return(image->size());
}

void Data::saveImage(QString fileName) {

  image->save(fileName);
}

void Data::scaleImage(double scale) {

  QImage scaledImage = image->scaled((double)image->width() *  scale, (double)image->height() * scale, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);
  *image = scaledImage;
} 

void Data::initChunkList() {

  int i;
  
  for (i = 0; i < imageHeight / chunkLen; i++) 
    chunkList.append(i * chunkLen);
}

int mandelPixel(double c_re, double c_im) {

  int iteration;
  double z_re1, z_re2, z_im1, z_im2, b;
  
  z_re1 = 0;
  z_im1 = 0;
  z_re2 = 0;
  z_im2 = 0;
  b = 0;
  iteration = 0;
  while ((b < 100.0) && (iteration < maxIterations)) {
    z_re2 = z_re1 * z_re1 - z_im1 * z_im1 + c_re;
    z_im2 = 2 * z_re1 * z_im1 + c_im;
    b = z_re2 * z_re2 + z_im2 * z_im2;
    z_re1 = z_re2;
    z_im1 = z_im2;
    iteration++;
  }   
  return(iteration);
}

QImage calcChunk(const int chunk) {

  int i1, i2;
  double z_re, z_im, xScale, yScale, xOfs, yOfs;
  QImage chunkImage(imageWidth, chunkLen, QImage::Format_RGB32);
  
  xOfs = -(double)imageWidth/1.25;
  yOfs = -(double)imageHeight/2.0;
  xScale = 2.6 / (double)imageWidth;
  yScale = 2.6 / (double)imageHeight;
  for (i1 = 0; i1 < chunkLen; i1++) 
    for (i2 = 0; i2 < imageWidth; i2++) {
      z_re = (double)(i2 + xOfs) * xScale;
      z_im = (double)(chunk + i1 + yOfs) * yScale;
      chunkImage.setPixel(i2, i1, palette[mandelPixel(z_re, z_im)].rgb());
    }
  return(chunkImage);
}

void collectChunks(QImage &mandelImage, const QImage &chunkImage) {

   static int offset = 0;
   if (!offset) mandelImage = QImage(imageWidth, imageHeight, QImage::Format_RGB32);
   QPainter p(&mandelImage);
   p.drawImage(QPoint(0, offset), chunkImage);
   offset += chunkLen;
}

void Data::calcMandelBrot() {

  mandelWatcher = new QFutureWatcher<QImage>; 
  QObject::connect(mandelWatcher, SIGNAL(finished()), this, SLOT(mandelFinished()));
  QObject::connect(mandelWatcher, SIGNAL(progressValueChanged(int)), this, SLOT(mandelProgress(int)));
  mandelImage = QtConcurrent::mappedReduced(chunkList, calcChunk, collectChunks, QtConcurrent::OrderedReduce);
  mandelWatcher->setFuture(mandelImage);
}

void Data::mandelFinished() {

  *image = mandelImage.result();
  emit(mandelReady());
}

void Data::mandelProgress(int value) {

   emit(mandelStatus(100 * value / mandelWatcher->progressMaximum()));
}
